/* ###
 * IP: GHIDRA
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
//Decompile an entire program
// Modified from Decompile.java, included with Ghidra
// Ben Lincoln, 2019

import java.io.*;
import java.util.*;

import ghidra.app.decompiler.*;
import ghidra.app.plugin.core.script.Ingredient;
import ghidra.app.plugin.core.script.IngredientDescription;
import ghidra.app.script.GatherParamPanel;
import ghidra.app.script.GhidraScript;
import ghidra.app.util.Option;
import ghidra.app.util.exporter.CppExporter;
import ghidra.program.model.address.*;
import ghidra.program.model.data.*;
import ghidra.program.model.listing.*;
import ghidra.program.model.lang.Register;
import ghidra.program.model.lang.RegisterManager;
import ghidra.program.model.mem.Memory;
import ghidra.program.model.mem.*;
import ghidra.program.model.symbol.RefType;
import ghidra.program.model.symbol.*;

import utilities.util.FileUtilities;

public class TDRDecompile extends GhidraScript implements Ingredient {

	@Override
	public void run() throws Exception {
		IngredientDescription[] ingredients = getIngredientDescriptions();
		for (IngredientDescription ingredient : ingredients) {
			state.addParameter(ingredient.getID(), ingredient.getLabel(), ingredient.getType(),
				ingredient.getDefaultValue());
		}
		if (!state.displayParameterGatherer("Script Options")) {
			return;
		}
		File outputFile = (File) state.getEnvironmentVar("COutputFile");
		/* CppExporter cppExporter = new CppExporter();
		List<Option> options = new ArrayList<Option>();
		options.add(new Option(CppExporter.CREATE_HEADER_FILE, new Boolean(true)));
		cppExporter.setOptions(options);
		cppExporter.setExporterServiceProvider(state.getTool());
		cppExporter.export(outputFile, currentProgram, null, monitor); */
		
		DecompileOptions options = new DecompileOptions();
		DecompInterface ifc = new DecompInterface();
		ifc.setOptions(options);
        
		if ( !ifc.openProgram(this.currentProgram) ) {
			throw new DecompileException("Decompiler", "Unable to initialize: "+ifc.getLastMessage());
		}
		/* ifc.setSimplificationStyle("normalize"); */
		FunctionIterator funcIter = currentProgram.getFunctionManager().getFunctions(true);
		
		FileWriter output = new FileWriter(outputFile);
		BufferedWriter writer = new BufferedWriter(output);
		
		while (funcIter.hasNext() && !monitor.isCancelled())
		{
			String currentTaskDescription = "";
			try
			{
				String functionName = "";
				String decompiledCode = "";
				String funcSig = "";
				boolean gotCode = true;
				currentTaskDescription = "while getting next function";
				Function func = funcIter.next();
				currentTaskDescription = "getting function name";
				functionName = func.getName();
				currentTaskDescription = "getting function entrypoint for " + functionName;
				Address funcAddress = func.getEntryPoint();
				currentTaskDescription = "getting function entrypoint offset for " + functionName;
				long funcOffset = funcAddress.getOffset();
				currentTaskDescription = "getting function " + functionName + " entrypoint offset " + String.valueOf(funcOffset) + " as hex string";
				String formattedOffset = Long.toHexString(funcOffset);
				currentTaskDescription = "getting DecompileResults for function " + functionName + " at " + formattedOffset;
				DecompileResults res = ifc.decompileFunction(func, 0, null);
				currentTaskDescription = "getting DecompiledFunction for function " + functionName + " at " + formattedOffset;
				DecompiledFunction df = res.getDecompiledFunction();
				currentTaskDescription = "getting C code for function " + functionName + " at " + formattedOffset;
				try
				{
					decompiledCode = df.getC();
				}
				catch (Exception e)
				{
					HandleDecompileException(e, currentTaskDescription);
					decompiledCode = "// " + e.toString() + " thrown " + currentTaskDescription + ": " + e.getMessage();
					gotCode = false;
				}
				if (gotCode)
				{
					currentTaskDescription = "getting signature for function at " + formattedOffset;
					try
					{
						funcSig = df.getSignature();
					}
					catch (Exception e)
					{
						HandleDecompileException(e, currentTaskDescription);
						funcSig = "// " + e.toString() + " thrown " + currentTaskDescription + ": " + e.getMessage();
						gotCode = false;
					}
				}
				currentTaskDescription = "writing function at " + formattedOffset + " to output file";
				if (gotCode)
				{
					writer.write("// BEGIN FUNCTION: " + funcSig + " @" + formattedOffset);
					writer.newLine();
					//writer.newLine();
				}
				writer.write(decompiledCode);
				writer.newLine();
				if (gotCode)
				{
					writer.write("// END FUNCTION");
					writer.newLine();
					//writer.newLine();
				}
			}
			catch (Exception e)
			{
				HandleDecompileException(e, currentTaskDescription);
			}
		}
		
		writer.close();
	}
	
	public void HandleDecompileException(Exception e, String taskDescription)
	{
		println(e.toString() + " thrown " + taskDescription + ": " + e.getMessage());
	}

	@Override
	public IngredientDescription[] getIngredientDescriptions() {
		IngredientDescription[] retVal = new IngredientDescription[] {
			new IngredientDescription("COutputFile", "Output C File", GatherParamPanel.FILE, "") };
		return retVal;
	}

}
